import os

from loguru import logger
from smart_open import open

from .glob import glob


def _aws_exists(path: str) -> bool:
    """Test if an AWS file exists."""
    try:
        from botocore.exceptions import ClientError  # pants: no-infer-dep
    except ImportError:
        logger.error(
            "Didn't find AWS dependencies, "
            "try installing image_tiles[aws] if you haven't already."
        )
        return False

    file_exists = False

    # Check for individual files.
    try:
        with open(path):
            file_exists = True
    except OSError:
        pass

    # Check for "folder", which means that folder has items
    # since S3 cannot have empty folders.
    try:
        items = glob(os.path.join(path, "*"))
        if items:
            file_exists = True
    except ClientError:
        pass

    return file_exists


def _gcs_exists(path: str) -> bool:
    """Test if a GCS file exists."""
    raise NotImplementedError


def _local_exists(path: str) -> bool:
    """Test if a local file exists."""
    return os.path.exists(path)


def exists(path: str) -> bool:
    """Check for existance of a file, maybe in the cloud.

    Args:
        path: An S3 path including the s3://

    Returns:
        exists: File exists on S3 at path given.
    """
    if "s3://" in path:
        return _aws_exists(path)
    elif "gs://" in path:
        return _gcs_exists(path)
    else:
        return _local_exists(path)



import unittest
from typing import Sequence

import boto3
import moto
from parameterized import parameterized

from .glob import _aws_glob, _filter_sequences


def index_from_lists(a: Sequence, b: Sequence) -> Sequence:
    """Return a list of indexes if an element of a is in b."""
    indexes = []
    for idx, item in enumerate(b):
        if item in a:
            indexes.append(idx)

    return indexes


class TestGlob(unittest.TestCase):
    @parameterized.expand(
        [
            ("path/to/file/*.png", [0, 1, 2, 3]),
            ("path/to/file/*.jpg", [4, 5]),
            ("path/to/file/*.txt", [6]),
            ("path/to/file/cat*.png", [1, 3]),
            ("path/to/file/cat*", [1, 3, 5, 6]),
        ]
    )
    def test_filter_sequences(self, pattern, result_indexes):
        possible_results = (
            "path/to/file/001.png",
            "path/to/file/cat.png",
            "path/to/file/dog.png",
            "path/to/file/catdog.png",
            "path/to/file/dog.jpg",
            "path/to/file/catdog.jpg",
            "path/to/file/catdog.txt",
        )

        results = _filter_sequences(pattern, possible_results)
        indexes = index_from_lists(results, possible_results)
        self.assertEqual(indexes, result_indexes)

    @moto.mock_s3
    def test_s3_empty_glob(self):
        conn = boto3.resource("s3", region_name="us-east-1")
        conn.create_bucket(Bucket="path")

        # Test an empty bucket
        results = _aws_glob("s3://path/*.png")
        self.assertEqual(results, [])

    @parameterized.expand(
        [
            ("s3://path/*.png", [0, 1, 2, 3]),
            ("s3://path/*.jpg", [4, 5]),
            ("s3://path/*.txt", [6]),
            ("s3://path/cat*.png", [1, 3]),
            ("s3://path/cat*", [1, 3, 5, 6]),
        ]
    )
    def test_s3_glob(self, pattern, result_indexes):
        with moto.mock_s3():
            conn = boto3.resource("s3", region_name="us-east-1")
            bucket_name = "path"
            conn.create_bucket(Bucket=bucket_name)

            # Add a bunch of files
            files = (
                "s3://path/001.png",
                "s3://path/cat.png",
                "s3://path/dog.png",
                "s3://path/catdog.png",
                "s3://path/dog.jpg",
                "s3://path/catdog.jpg",
                "s3://path/catdog.txt",
            )
            s3 = boto3.client("s3", region_name="us-east-1")
            for f in files:
                filename = f.split("/")[-1]
                s3.put_object(Bucket=bucket_name, Key=filename, Body="nodata")

            results = _aws_glob(pattern)
            indexes = index_from_lists(results, files)
            self.assertEqual(indexes, result_indexes)


if __name__ == "__main__":
    unittest.main()



"""Provides normalization functions for images in np.array format."""

from functools import partial
from typing import Callable, Literal

import numpy as np
from numpy.typing import ArrayLike


def _format_array(image: np.ndarray, batch_operation: str, channel_operation: str):
    """Formats an array to work with the internals of the normalization
    functions. This means putting the array into channel_operation-last format
    and including a batch dimension (even if that dimension is only 1).

    Args:
        array: The input array to reformat

    Returns:
        reformatted: A reformatted array
    """
    if batch_operation == "expand":
        image = np.expand_dims(image, 0)

    if channel_operation == "to_first":
        image = np.einsum("bwhc->bcwh", image)
    elif channel_operation == "to_last":
        image = np.einsum("bcwh->bwhc", image)
    else:
        pass

    if batch_operation == "squeeze":
        image = np.squeeze(image)
    return image


def _get_format_fx(
    image: np.ndarray, image_channel_format: Literal["last", "first"]
) -> tuple[Callable, Callable]:
    add_batch = len(image.shape) == 3
    convert_ordering = image_channel_format == "first"

    apply_batch = "expand" if add_batch else None
    apply_ordering = "to_last" if convert_ordering else None
    apply_fx = partial(
        _format_array, batch_operation=apply_batch, channel_operation=apply_ordering
    )

    revert_batch = "squeeze" if add_batch else None
    revert_ordering = "to_first" if convert_ordering else None
    revert_fx = partial(
        _format_array, batch_operation=revert_batch, channel_operation=revert_ordering
    )

    return apply_fx, revert_fx


def standard_normalization(image: np.ndarray) -> np.ndarray:
    """If the image is a standard format, leave it alone. Otherwise apply
    scaling norm.

    Args:
        image: An input image to normalize, any scale and format

    Returns:
        scaled: An image in uint8 format.
    """
    _, _, c = image.shape
    if c in [1, 3, 4]:
        if image.dtype in [np.float32, np.uint8]:
            return image

    return scaling_normalization(image)


def scaling_normalization(image: np.ndarray) -> np.ndarray:
    """Normalize an image by scaling to uint8.

    This function will clip numbers under 0 and scale postive numbers.

    Args:
        image: An input image to normalize, any scale and format

    Returns:
        scaled: An image in uint8 format.
    """
    # Convert to float for scaling
    image = image.astype(float)
    # Clip negative numbers, but leave small positive numbers alone.
    image[image <= 0] = 0
    # Scale from 0 to 1
    image /= np.max(image)
    # Scale from 0 to 255
    image *= 255

    return image.astype(np.uint8)


def sigmoid_normalization(
    image: np.ndarray,
    c: float = 10.0,
    th: float = 0.125,
    channels: Literal["last", "first"] = "last",
) -> np.ndarray:
    """Normalize an image by a sigmoid function like in xarray-spatial. The
    function accepts both HWC and BHWC arrays.

    Normalize by the function
    ``normalized_pixel = 1 / (1 + np.exp(c * (th - normalized_pixel)))``

    Args:
        image: An input image to normalize.
        c: Contrast controlling parameter.
        th: Brightness controlling parameter.
    """
    apply_fx, revert_fx = _get_format_fx(image, channels)
    image = apply_fx(image)

    max_pixel_val = 255

    min_val = np.nanmin(image, axis=(1, 2, 3))
    max_val = np.nanmax(image, axis=(1, 2, 3))
    range_val = np.maximum(max_val - min_val, 1)

    norm = (image - min_val.reshape((-1, 1, 1, 1))) / range_val.reshape((-1, 1, 1, 1))
    # sigmoid contrast enhancement
    norm = 1 / (1 + np.exp(c * (th - norm)))
    normalized = norm * max_pixel_val

    return revert_fx(normalized).astype(np.uint8)


def sentinel_truecolor_image(image: np.ndarray, normalizer: int = 2000) -> np.ndarray:
    """Convert a sentinel l2a or l1a product to a true color image, using
    published values of maximum reflectance.

    Args:
        image: An unnormalized sentinel-2 image.
        normalizer: The value to be used for 255, matches published TCI algo.

    Returns:
        rgb: A true-color image of the input.
    """
    assert image.shape[2] == 3, "Expect a 3 channel RGB image"

    normalized_image = np.array(image)
    saturated_idx_flat = (normalized_image > normalizer).max(axis=2)
    saturated_idx = np.repeat(np.expand_dims(saturated_idx_flat, 2), 3, axis=2)

    normalized_image = image / float(normalizer)
    rgb = normalized_image * 255
    rgb[saturated_idx] = 255
    return (rgb).astype(np.uint8)


def invert_mean_std_normalization(
    image: np.ndarray,
    mean: ArrayLike,
    std: ArrayLike,
    channels: Literal["last", "first"] = "last",
) -> np.ndarray:
    """Invert an image normalized by a set mean and standard deviation. The
    function accepts both HWC and BHWC arrays.

    Args:
        image: An input image to normalize.
        c: Contrast controlling parameter.
        th: Brightness controlling parameter.
    """
    mean = np.expand_dims(mean, (0, 1, 2))
    std = np.expand_dims(std, (0, 1, 2))

    apply_fx, revert_fx = _get_format_fx(image, channels)
    image = apply_fx(image)
    normalized = (image * std) + mean

    return revert_fx(normalized)



import argparse
import io
import os
import random
from typing import Optional

import flask
import imageio.v3 as imageio
from loguru import logger

from .file_exists import exists
from .glob import glob
from .image import get_supported_extensions, read_image


def parse_image_tile_args():
    parser = argparse.ArgumentParser(
        description="Simple server visualizes images in a folder."
    )
    parser.add_argument("folder_path", help="Which folder to serve.")
    parser.add_argument(
        "--bind",
        help="Which address to bind to, 0.0.0.0 will bind to all IPs.",
        default="0.0.0.0",
    )
    parser.add_argument("--port", help="Port to bind on.", type=int, default=8000)
    parser.add_argument(
        "--render_method",
        default="rgb",
        nargs="?",
        choices=["rgb", "bgr", "bw", "sentinel"],
        help="Rendering method, see README for details.",
    )
    parser.add_argument(
        "--normalization",
        default="standard",
        nargs="?",
        choices=["standard", "scaling", "sigmoid", "sentinel"],
        help="Normalization method, see README for details.",
    )
    parser.add_argument(
        "--debug", help="Enable server Flask debugging.", type=bool, default=False
    )
    parser.add_argument(
        "--root",
        help="Root directory to serve templates/static from.",
        type=str,
        default="",
    )
    parser.add_argument(
        "--num_items", help="Maximum images to show on the page.", type=int, default=100
    )

    args = parser.parse_args()

    # Do some surgery for relative paths
    if not args.folder_path.startswith("/"):
        if "s3://" not in args.folder_path and "gs://" not in args.folder_path:
            args.folder_path = os.path.join(os.getcwd(), args.folder_path)

    return args


def get_image_files(folder: str) -> list[str]:
    """Get a list of supported image files in a folder."""
    supported_extensions = get_supported_extensions()

    # We grab the filelist once and then test if a file is valid. This
    # costs more memory but the cost of a large glob in the cloud is high.
    path = os.path.join(folder, "*")
    logger.info("Globbing files, this may take a while for a large AWS/GCP bucket...")
    filelist = glob(path)

    imagelist = []
    for f in filelist:
        ext = "." + f.split(".")[-1]
        if ext in supported_extensions:
            imagelist.append(f)

    logger.info(f"Got {len(imagelist)} items from {folder}.")
    return imagelist


def read_and_render_image(path: str, args: argparse.Namespace) -> Optional[bytes]:
    """Reads an image and converts it to a png for display in the web
    browser."""
    # Rewrite based on the input
    if path.startswith("/s3"):
        path = path.replace("/s3/", "s3://")
    elif path.startswith("/gs"):
        path = path.replace("/gs/", "gs://")

    if not exists(path):
        return None

    array = read_image(
        path, normalize=args.normalization, render_method=args.render_method
    )

    with io.BytesIO() as f:
        imageio.imwrite(uri=f, image=array, extension=".png")  # type: ignore
        return f.getvalue()


def server_path(ext_path: str) -> str:
    """Convert an external path to one we can use internally."""
    if "s3://" in ext_path:
        return f"images/s3/{ext_path[5:]}"
    elif "gs://" in ext_path:
        return f"images/{ext_path[3:]}"
    else:
        return f"images/{ext_path[1:]}"


def flask_app(args):
    """Defines the flask app itself."""
    root = os.path.join(os.getcwd(), args.root) if args.root else ""
    app = flask.Flask(
        __name__,
        template_folder=os.path.join(root, "templates"),
        static_folder=os.path.join(root, "static"),
    )

    @app.route("/images/<path:image_path>")
    def image(image_path):
        full_path = f"/{image_path}"
        data = read_and_render_image(full_path, args)

        return flask.Response(data, mimetype="image/png")

    @app.route("/")
    def index():
        url_args = flask.request.args

        num_items = args.num_items
        if "num_items" in url_args:
            num_items = int(url_args["num_items"])

        logger.info(f"Reading from folder {args.folder_path}")
        filelist = get_image_files(args.folder_path)
        random.shuffle(filelist)
        filelist = filelist[0:num_items]
        items = [server_path(f) for f in filelist]
        logger.info(f"Request recieved, found {len(filelist)} items to display")

        return flask.render_template(
            "image_tiles.html",
            folder_location=args.folder_path,
            num_images=num_items,
            items=items,
            normalization=args.normalization,
            render_method=args.render_method,
        )

    return app


def image_tile_server():
    args = parse_image_tile_args()
    logger.info("Starting image_tiles server on http://%s:%s" % (args.bind, args.port))
    app = flask_app(args)
    app.run(host=args.bind, port=args.port, debug=args.debug, use_reloader=False)


if __name__ == "__main__":
    image_tile_server()



"""Provides image utility functions."""

import io
from typing import Callable, Optional

import imageio
import numpy as np
import tifffile
from PIL import Image
from smart_open import open

from .normalization import (
    scaling_normalization,
    sentinel_truecolor_image,
    sigmoid_normalization,
    standard_normalization,
)


def _render_image_data(data: np.ndarray, render_method: str) -> np.ndarray:
    """Convert a numpy array into an image we can actually render."""
    if len(data.shape) == 2:
        # If it's one channel expand the dimensions and make it a grayscale image
        data = np.expand_dims(data, -1)
    elif data.shape[2] >= 3:
        if render_method == "rgb":
            # For RGB data, take the first 3 channels (ignore alpha)
            data = data[:, :, 0:3]
        elif render_method == "bgr":
            # For BGR data, we'll do the same as above but also invert the axes
            data = data[:, :, 0:3]
            data = data[:, :, ::-1]
        elif render_method == "sentinel":
            # For satellite data from Sentinel, we'll take 3 spectral bands
            data = data[:, :, 1:4]
            data = data[:, :, ::-1]
        elif render_method == "bw":
            # Apply greyscale to this RGB image
            data = 0.30 * data[:, :, 0] + 0.59 * data[:, :, 1] + 0.11 * data[:, :, 2]
            data = np.expand_dims(data, -1)
        else:
            raise ValueError(
                f"Not a valid type of image rendering, got {render_method}"
            )
    else:
        # If the dimensions are <2 and >3 then we have some weird tensor.
        raise ValueError(f"Image dimensionality for render is wrong ({data.shape})")

    return data


def get_supported_extensions() -> set:
    """Get a list of supported extensions for this module, this is all PIL
    extensions plus our own.

    Args: None

    Returns:
        supported_extensions: A set of possible extensions.
    """
    exts = Image.registered_extensions()
    supported_extensions = {ex for ex, f in exts.items() if f in Image.OPEN}
    additional_extensions = {".tif", ".tiff"}
    return supported_extensions.union(additional_extensions)


def read_image(
    path: str, normalize: Optional[str] = "sigmoid", render_method: str = "rgb"
) -> np.ndarray:
    """Read an image file, possibly containing many channels.

    Args:
        path: A filesystem path to the file
        normalize: How to normalize the image, if at all, one of
            {None, sigmoid, sentinel}
        render_method: Which multichannel image format to use, one
            of {bw, rgb, bgr, sentinel}

    Returns:
        raw_bytes: A bytestream containing the jpeg image
    """
    # Normalize functions supported.
    normalize_fx: dict[str, Callable] = {
        "standard": standard_normalization,
        "scaling": scaling_normalization,
        "sigmoid": sigmoid_normalization,
        "sentinel": sentinel_truecolor_image,
    }

    # Reader functions supported, but we'll fallback to PIL if we don't
    # have a specialized function here.
    format = path.split(".")[-1]
    reader_fx = {
        "tiff": tifffile.imread,
        "tif": tifffile.imread,
    }

    # Select an image reader, or fallback to PIL for anything else.
    reader = imageio.imread
    if format in reader_fx.keys():
        reader = reader_fx[format]

    # Read our image and render it into RGB for the webpage
    with open(path, "rb") as f:
        file_bytes = io.BytesIO(f.read())
        data = np.asarray(reader(file_bytes))
        rendered_data = _render_image_data(data, render_method=render_method)

        if normalize is not None:
            rendered_data = normalize_fx[normalize](rendered_data)

    return rendered_data



import unittest

import boto3
import moto
from parameterized import parameterized

from .file_exists import _aws_exists, exists


class Test(unittest.TestCase):
    @parameterized.expand(
        [
            ("s3://path/cat.png", True),
            ("s3://path/koala.png", False),
            ("s3://newpath/cat.png", False),
        ]
    )
    def test_aws_exists(self, path, result):
        with moto.mock_s3():
            conn = boto3.resource("s3", region_name="us-east-1")
            bucket_name = "path"
            conn.create_bucket(Bucket=bucket_name)

            # Add a bunch of files
            files = (
                "s3://path/001.png",
                "s3://path/cat.png",
                "s3://path/dog.png",
                "s3://path/catdog.png",
                "s3://path/dog.jpg",
                "s3://path/catdog.jpg",
                "s3://path/catdog.txt",
            )
            s3 = boto3.client("s3", region_name="us-east-1")
            for f in files:
                filename = f.split("/")[-1]
                s3.put_object(Bucket=bucket_name, Key=filename, Body="nodata")

            exist = _aws_exists(path)
            self.assertEqual(exist, result)

            exist = exists(path)
            self.assertEqual(exist, result)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
from parameterized import parameterized

from .image import _render_image_data, get_supported_extensions


class TestGetSupportedExtensions(unittest.TestCase):
    def test_supported_ext(self):
        supported = get_supported_extensions()
        self.assertTrue(".tif" in supported)
        self.assertTrue(".tiff" in supported)
        self.assertTrue(len(supported) > 2)


class TestRenderImageData(unittest.TestCase):
    @parameterized.expand(
        [("rgb", 480), ("bgr", 512), ("sentinel", 528), ("bw", 492.96)]
    )
    def test_rendering(self, render_type, answer):
        test_array = np.arange(16 * 16 * 4).reshape((16, 16, 4))
        rendered = _render_image_data(test_array, render_type)
        result = rendered[0, :, 0].sum()

        self.assertAlmostEqual(result, answer)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from .normalization import (
    invert_mean_std_normalization,
    scaling_normalization,
    sentinel_truecolor_image,
    sigmoid_normalization,
    standard_normalization,
)


class TestScalingNormalization(unittest.TestCase):
    def test_scaling_image(self):
        test_array = np.arange(16 * 16 * 3).reshape((16, 16, 3))

        normed = scaling_normalization(test_array)
        self.assertAlmostEqual(normed.sum(), 97537)
        self.assertTrue(normed.max(), 255)
        self.assertEqual(normed.dtype, np.uint8)


class TestStandardNormalization(unittest.TestCase):
    def test_good_image(self):
        test_array = np.ones(16 * 16 * 3).reshape((16, 16, 3))
        test_array = test_array.astype(float)

        normed = standard_normalization(test_array)
        self.assertAlmostEqual(normed.sum(), 16 * 16 * 3 * 255)
        self.assertTrue(normed.max(), 255)
        self.assertEqual(normed.dtype, np.uint8)

    def test_bad_image(self):
        test_array = np.arange(16 * 16 * 3).reshape((16, 16, 3))

        normed = standard_normalization(test_array)
        self.assertAlmostEqual(normed.sum(), 97537)
        self.assertTrue(normed.max(), 255)
        self.assertEqual(normed.dtype, np.uint8)


class TestSentinelTrueColor(unittest.TestCase):
    def test_sentinel_norm(self):
        test_array = np.arange(16 * 16 * 3).reshape((16, 16, 3))
        normed = sentinel_truecolor_image(test_array)
        self.assertAlmostEqual(normed.sum(), 37170)
        self.assertTrue(normed.max(), 255)
        self.assertEqual(normed.dtype, np.uint8)


class TestSigmoidNormalization(unittest.TestCase):
    def test_sig_norm(self):
        test_array = np.arange(16 * 16 * 3).reshape((16, 16, 3))
        normed = sigmoid_normalization(test_array, 10, 0.125)
        self.assertAlmostEqual(normed.sum(), 165926)
        self.assertTrue(normed.max(), 255)
        self.assertEqual(normed.dtype, np.uint8)

        normed = sigmoid_normalization(test_array, 5, 0.25)
        self.assertAlmostEqual(normed.sum(), 137511)
        self.assertTrue(normed.max(), 255)
        self.assertEqual(normed.dtype, np.uint8)

    def test_sig_norm_batch(self):
        batch_test_array = np.arange(8 * 16 * 16 * 3).reshape((8, 16, 16, 3))
        normed = sigmoid_normalization(batch_test_array, 10, 0.125)
        self.assertAlmostEqual(normed.sum(), 1327408)
        self.assertTrue(normed.max(), 255)
        self.assertEqual(normed.dtype, np.uint8)

    def test_sig_norm_channels_first(self):
        batch_test_array = np.arange(16 * 16 * 3).reshape((3, 16, 16))
        normed = sigmoid_normalization(batch_test_array, 10, 0.125, channels="first")
        self.assertAlmostEqual(normed.sum(), 165926)
        self.assertTrue(normed.max(), 255)
        self.assertEqual(normed.dtype, np.uint8)

    def test_sig_norm_channels_first_batch(self):
        batch_test_array = np.arange(8 * 16 * 16 * 3).reshape((8, 3, 16, 16))
        normed = sigmoid_normalization(batch_test_array, 10, 0.125, channels="first")
        self.assertAlmostEqual(normed.sum(), 1327408)
        self.assertTrue(normed.max(), 255)
        self.assertEqual(normed.dtype, np.uint8)


class TestMeanStdNormalization(unittest.TestCase):
    def test_mean_std_norm(self):
        test_array = np.arange(16 * 16 * 3).reshape((16, 16, 3))
        mean = np.array((0.485, 0.456, 0.406))
        std = np.array((0.229, 0.224, 0.225))
        normed = invert_mean_std_normalization(test_array, mean=mean, std=std)
        self.assertAlmostEqual(normed.sum(), 66907.136)

    def test_mean_std_norm_batch(self):
        test_array = np.arange(8 * 16 * 16 * 3).reshape((8, 16, 16, 3))
        mean = (0.485, 0.456, 0.406)
        std = (0.229, 0.224, 0.225)
        normed = invert_mean_std_normalization(test_array, mean=mean, std=std)
        self.assertAlmostEqual(normed.sum(), 4267663.36)

    def test_mean_std_norm_channels_first(self):
        test_array = np.arange(16 * 16 * 3).reshape((3, 16, 16))
        mean = (0.485, 0.456, 0.406)
        std = (0.229, 0.224, 0.225)
        normed = invert_mean_std_normalization(
            test_array, mean=mean, std=std, channels="first"
        )
        self.assertAlmostEqual(normed.sum(), 66646.016)

    def test_mean_std_norm_channels_first_batch(self):
        test_array = np.arange(8 * 16 * 16 * 3).reshape((8, 3, 16, 16))
        mean = (0.485, 0.456, 0.406)
        std = (0.229, 0.224, 0.225)
        normed = invert_mean_std_normalization(
            test_array, mean=mean, std=std, channels="first"
        )
        self.assertAlmostEqual(normed.sum(), 4265574.4)


if __name__ == "__main__":
    unittest.main()



import fnmatch
import glob as sys_glob
import os
from typing import Sequence

from loguru import logger


def _filter_sequences(pattern: str, search_list: Sequence) -> list:
    """Return only the items from search_list that match pattern."""

    def match_file(filename):
        return fnmatch.fnmatch(filename, pattern)

    return list(filter(match_file, search_list))


def _aws_glob(path: str) -> list:
    """Implementation of AWS glob."""
    try:
        import boto3
    except ImportError:
        logger.error(
            "Didn't find AWS dependencies, "
            "try installing image_tiles[aws] if you haven't already."
        )
        return []

    # Do a bunch of surgery on the path to get the components
    full_path, _ = os.path.split(path)
    _, _, bucket, *remaining = full_path.split("/")
    prefix = "/".join(remaining)

    # For wildcard folder globs (i.e. /path/to/**/*.jpg) we need to do a
    # bit more processing.
    prefix = prefix.replace("/**", "")

    # List all the objects in the path via pagination since there could
    # be more than 1000 objects in our bucket
    paginator = boto3.client("s3").get_paginator("list_objects_v2")
    pages = paginator.paginate(Bucket=bucket, Prefix=prefix)

    filenames = []
    for page in pages:
        # Construct the filenames from the results
        if "Contents" in page:
            keys = [f"s3://{bucket}/{o['Key']}" for o in page["Contents"]]

            # Add to our results
            filenames.extend(keys)

    # Do glob filtering
    return _filter_sequences(path, filenames)


def _gcs_glob(path: str) -> list:
    """Implementation of GCS glob."""
    raise NotImplementedError


def _local_glob(path: str) -> list:
    """Implementation of local glob."""
    return sys_glob.glob(path)


def glob(path: str) -> list:
    """Glob with support for cloud services.

    Args:
        path: An input path, includes the uri like s3:// or gs:// if in the cloud.

    Returns:
        results: A list of files returned from the glob.
    """
    if "s3://" in path:
        return _aws_glob(path)
    elif "gs://" in path:
        return _gcs_glob(path)
    else:
        return _local_glob(path)
